//	File: TextEncoding.cpp

//	EUC-jp <-> JIS conversion routine based on public domain code by Ken Lunde

#include "TextEncoding.h"

#define EUC         4
#define SJIS        5
#define SS2         142
#define ISEUC(A)    (A >= 161 && A <= 254)
#define HANKATA(A)  (A >= 161 && A <= 223)
#define ISNIGORI(A) ((A >= 182 && A <= 196) || (A >= 202 && A <= 206) || (A == 179))
#define ISMARU(A)   (A >= 202 && A <= 206)

static bool SkipESCSeq(int32* pos, char temp, bool* intwobyte);
static void seven2euc(const char* src, int32* srcLen, char* dst, int32* dstLen);
static void euc2seven(const char* src, int32* srcLen, char* dst, int32* dstLen, char ki[], char ko[]);
static void han2zen(const char* src, int32* srcLen, int* p1, int* p2, int incode);
static void sjis2jis(int* p1, int* p2);

status_t
ConvertToUTF8(uint32 srcEncoding,
			  const char* src,
			  int32* srcLen,
			  char* dst,
			  int32* dstLen,
			  int32* state,
			  char substitute)
{
	if (srcEncoding != B_JIS_CONVERSION)
		return convert_to_utf8(srcEncoding, src, srcLen, dst, dstLen, state, substitute);

	status_t aErr = B_OK;

	int32 aLengthEUC = *srcLen;
	char* aTextEUC = new char[aLengthEUC];
	aLengthEUC -= 1;

	seven2euc(src, srcLen, aTextEUC, &aLengthEUC);
	aTextEUC[aLengthEUC] = '\0';

#if (defined(B_BEOS_VERSION) && (B_BEOS_VERSION >= 0x0400))
//	int32 aState = 0;
	aErr = convert_to_utf8(B_EUC_CONVERSION, aTextEUC, &aLengthEUC, dst, dstLen, state, '?');
#else
	aErr = convert_to_utf8(B_EUC_CONVERSION, aTextEUC, &aLengthEUC, dst, dstLen);
#endif

	delete[] aTextEUC;

	return aErr;
}


status_t
ConvertFromUTF8(uint32 dstEncoding,
				const char* src,
				int32* srcLen,
				char* dst,
				int32* dstLen,
				int32* state,
				char substitute)
{
	if (dstEncoding != B_JIS_CONVERSION)
		return convert_from_utf8(dstEncoding, src, srcLen, dst, dstLen, state, substitute);

	status_t aErr;

	int32 aLengthEUC = *dstLen;
	char* aTextEUC = new char[aLengthEUC];
	aLengthEUC -= 1;

#if (defined(B_BEOS_VERSION) && (B_BEOS_VERSION >= 0x0400))
//	int32 aState = 0;
	aErr = convert_from_utf8(B_EUC_CONVERSION, src, srcLen, aTextEUC, &aLengthEUC, state, '?');
#else
	aErr = convert_from_utf8(B_EUC_CONVERSION, src, srcLen, aTextEUC, &aLengthEUC);
#endif
	aTextEUC[aLengthEUC] = '\0';

	if (aErr != B_OK)
		return aErr;

	euc2seven(aTextEUC, &aLengthEUC, dst, dstLen, "$B", "(B");

	delete[] aTextEUC;

	return aErr;
}


bool
SkipESCSeq(int32* pos, char temp, bool* intwobyte)
{
	bool aTemp = *intwobyte;

	if (temp == '$' || temp == '(')
		(*pos)++;

	if (temp == 'K' || temp == '$')
		*intwobyte = true;
	else
		*intwobyte = false;

	if (aTemp == *intwobyte)
		return false;
	else
		return true;
}


void
seven2euc(const char* src1, int32* srcLen, char* dst1, int32* dstLen)
{
//	char temp, p1, p2;
	int temp, p1, p2;
	bool aInTwoByte = false;
	int32 i = 0;
	int32 j = 0;

	const unsigned char* src = (unsigned char*)src1;
	unsigned char* dst = (unsigned char*)dst1;

	while (i < *srcLen && j < *dstLen) {
		p1 = src[i++];
		if (p1 == 0x1b) {
			temp = src[i++];
			SkipESCSeq(&i, temp, &aInTwoByte);
		} else if (p1 == 0x0a || p1 == 0x0d) {
			if (aInTwoByte)
				aInTwoByte = false;
			dst[j++] = p1;
		} else {
			if (aInTwoByte) {
				p2 = src[i++];
				p1 |= 0x80;
				p2 |= 0x80;
				dst[j++] = p1;
				dst[j++] = p2;
			} else {
				dst[j++] = p1;
			}
		}
	}
	*srcLen = i;
	*dstLen = j;
}


void
euc2seven(const char* src1, int32* srcLen, char* dst1, int32* dstLen, char ki[], char ko[])
{
	int incode = EUC;

//	char p1, p2;
	int p1, p2;
	bool intwobyte = false;

	const unsigned char* src = (unsigned char*)src1;
	unsigned char* dst = (unsigned char*)dst1;

	int32 i = 0;
	int32 j = 0;
	int32 k = 0;

	while (i < *srcLen && j < *dstLen) {
		p1 = src[i++];
		if (p1 == 0x0a || p1 == 0x0d) {
			if (intwobyte) {
				intwobyte = false;
				dst[j++] = 0x1b;
				for (k = 0; ko[k]; k++)
					dst[j++] = ko[k];
			}
			dst[j++] = p1;
		} else {
			if (ISEUC(p1)) {
				p2 = src[i++];
				if (ISEUC(p2)) {
					p1 &= 0x7f;
					p2 &= 0x7f;
					if (!intwobyte) {
						intwobyte = true;
						dst[j++] = 0x1b;
						for (k = 0; ki[k]; k++)
							dst[j++] = ki[k];
					}
				}
				dst[j++] = p1;
				dst[j++] = p2;
			} else if (p1 == SS2) {
				p2 = src[i++];
				if (HANKATA(p2)) {
					p1 = p2;
					han2zen(src1, &i, &p1, &p2, incode);
					sjis2jis(&p1, &p2);
					if (!intwobyte) {
						intwobyte = true;
						dst[j++] = 0x1b;
						for (k = 0; ki[k]; k++)
							dst[j++] = ki[k];
					}
				}
				dst[j++] = p1;
				dst[j++] = p2;
			} else {
				if (intwobyte) {
					intwobyte = false;
					dst[j++] = 0x1b;
					for (k = 0; ko[k]; k++)
						dst[j++] = ko[k];
				}
				dst[j++] = p1;
			}
		}
	}
	if (intwobyte) {
		dst[j++] = 0x1b;
		for (k = 0; ko[k]; k++)
			dst[j++] = ko[k];
	}
	*srcLen = i;
	*dstLen = j;
}


void
sjis2jis(int* p1, int* p2)
{
	unsigned char c1 = *p1;
	unsigned char c2 = *p2;
	bool adjust = (c2 < 159);
	int rowOffset = c1 < 160 ? 112 : 176;
	int cellOffset = adjust ? (c2 > 127 ? 32 : 31) : 126;

	if (adjust)
		*p1 = ((c1 - rowOffset) << 1) - 1;
	else
		*p1 = ((c1 - rowOffset) << 1);

	*p2 -= cellOffset;
}


void
han2zen(const char* src, int32* pos, int* p1, int* p2, int incode)
{
  int tmp = *p1,junk;
	bool maru = false;
	bool nigori = false;

  int mtable[][2] = {
    {129,66},{129,117},{129,118},{129,65},{129,69},{131,146},{131,64},
    {131,66},{131,68},{131,70},{131,72},{131,131},{131,133},{131,135},
    {131,98},{129,91},{131,65},{131,67},{131,69},{131,71},{131,73},
    {131,74},{131,76},{131,78},{131,80},{131,82},{131,84},{131,86},
    {131,88},{131,90},{131,92},{131,94},{131,96},{131,99},{131,101},
    {131,103},{131,105},{131,106},{131,107},{131,108},{131,109},
    {131,110},{131,113},{131,116},{131,119},{131,122},{131,125},
    {131,126},{131,128},{131,129},{131,130},{131,132},{131,134},
    {131,136},{131,137},{131,138},{131,139},{131,140},{131,141},
    {131,143},{131,147},{129,74},{129,75}
  };

	if (incode == SJIS) {
		*p2 = src[*pos++];
		if (*p2 == 222) {
			if (ISNIGORI(*p1))
				nigori = true;
			else
				pos--;
		} else if (*p2 == 223) {
			if (ISMARU(*p1))
				maru = true;
			else
				pos--;
		} else
			pos--;
	} else if (incode == EUC) {
		junk = src[*pos++];
		if (junk == SS2) {
			*p2 = src[*pos++];
			if (*p2 == 222) {
				if (ISNIGORI(*p1))
					nigori = true;
				else
					*pos -= 2;
			} else if (*p2 == 223) {
				if (ISMARU(*p1))
					maru = true;
				else
					*pos -= 2;
			} else
				*pos -= 2;
		} else
			*pos--;
	}
	if (*p1 >= 161 && *p1 <= 223) {
		*p1 = mtable[tmp - 161][0];
		*p2 = mtable[tmp - 161][1];
	}
	if (nigori) {
		if ((*p2 >= 74 && *p2 <= 103) || (*p2 >= 110 && *p2 <= 122))
			(*p2)++;
		else if (*p1 == 131 && *p2 == 69)
			*p2 = 148;
	} else if (maru && *p2 >= 110 && *p2 <= 122)
		*p2 += 2;
}
